/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2. */

#include "imspector.h"

#define PLUGIN_NAME "ACL IMSpector filter plugin"
#define PLUGIN_SHORT_NAME "ACL"

struct listelement
{
	std::string localid;
	std::vector<std::string> remoteids;
};

extern "C"
{
	bool initfilterplugin(struct filterplugininfo &filterplugininfo,
		class Options &options, bool debugmode);
	void closefilterplugin(void);
	bool eventfilter(std::vector<struct imevent> &imevents);
};

std::vector<struct listelement> whitelist;
std::vector<struct listelement> blacklist;
bool blockunlisted = false;
bool localdebugmode = false;

bool parselist(std::vector<struct listelement> &list, std::string listfilename);
void debuglist(std::vector<struct listelement> &list);
bool matchlist(std::string localid, std::string remoteid, std::vector<struct listelement> &list);

bool initfilterplugin(struct filterplugininfo &filterplugininfo,
	class Options &options, bool debugmode)
{
	localdebugmode = debugmode;
	
	filterplugininfo.pluginname = PLUGIN_NAME;
	
	std::string whitelistfilename = options["whitelist_filename"];
	std::string blacklistfilename = options["blacklist_filename"];
	
	if (options["block_unlisted"] == "on")
	{
		blockunlisted = true;
		debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Blacklisting unmatched messages");
	}
	
	if (whitelistfilename != "")
	{
		if (!(parselist(whitelist, whitelistfilename))) return false;
		debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Whitelist %s size: %d", 
			whitelistfilename.c_str(), whitelist.size());
		debuglist(whitelist);
	}
	if (blacklistfilename != "")
	{
		if (!(parselist(blacklist, blacklistfilename))) return false;
		debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Blacklist %s size: %d",
			blacklistfilename.c_str(), blacklist.size());
		debuglist(blacklist);
	}
	
	return true;
}

void closefilterplugin(void)
{
	return;
}

/* The main plugin function. See eventplugin.cpp. */
bool eventfilter(std::vector<struct imevent> &imevents)
{
	bool lastfiltered = false;
	
	for (std::vector<struct imevent>::iterator i = imevents.begin();
		i != imevents.end(); i++)
	{
		/* Look for whitelist match first, then blacklist. */
		if (matchlist((*i).localid, (*i).remoteid, whitelist))
			continue;
		
		if (matchlist((*i).localid, (*i).remoteid, blacklist))
		{
			(*i).filtered = true;
			lastfiltered = true;
			continue;
		}
		
		if (blockunlisted)
		{
			(*i).filtered = true;
			lastfiltered = true;
		}
	}
	
	return lastfiltered;
}

/* The real guts of this thing.  Returns true if a match is found. */
bool matchlist(std::string localid, std::string remoteid, std::vector<struct listelement> &list)
{
	debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Local: %s Remote: %s",
		localid.c_str(), remoteid.c_str());
	
	for (std::vector<struct listelement>::iterator i = list.begin();
		i != list.end(); i++)
	{
		if ((*i).localid == localid)
		{
			debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Got Local match");
			
			if (!(*i).remoteids.size())
			{
				debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Remote list empty; matching");
				return true;
			}		
			for (std::vector<std::string>::iterator j = (*i).remoteids.begin();
				j != (*i).remoteids.end(); j++)
			{
				if ((*j) == remoteid)
				{
					debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Got Remote match");
					return true;
				}
			}
			/* Local match, but remote match has ran off the end. */
			break;
		}
	}
	
	debugprint(localdebugmode, PLUGIN_SHORT_NAME ": No match");
	
	return false;
}

/* Parse in a white/black list. */
bool parselist(std::vector<struct listelement> &list, std::string listfilename)
{
	FILE *hfile = NULL;
	char buffer[STRING_SIZE];
	char *colon;
	char *remoteid;
	int c = 0;
	std::string temp;
	
	memset(buffer, 0, STRING_SIZE);
	
	if (!(hfile = fopen(listfilename.c_str(), "r")))
		return false;
		
	while (fgets(buffer, STRING_SIZE, hfile))
	{
		/* Remove the \n. */
		char *t = strchr(buffer, '\n');
		if (t) *t = '\0';		
        
		/* Ignore lines with no colon or comment lines. */
		if (buffer[0] == '#') continue;
		if (!(colon = strchr(buffer, ':'))) continue;

		*colon = '\0';
		
		struct listelement listelement;
	
		temp = buffer;
	
		std::transform(temp.begin(), temp.end(), temp.begin(), tolower);
		listelement.localid = temp;
		
		char *p = colon + 1;
		
		while ((remoteid = strsep(&p, " ")))
		{
			if (!(strlen(remoteid))) continue;
			
			temp = remoteid;
		
			std::transform(temp.begin(), temp.end(), temp.begin(), tolower);
			listelement.remoteids.push_back(temp);			
		}
		
		list.push_back(listelement);
		
		c++;
	}
	
	fclose(hfile);
	
	return c;
}

/* Output a list to the debug log. */
void debuglist(std::vector<struct listelement> &list)
{
	for (std::vector<struct listelement>::iterator i = list.begin();
		i != list.end(); i++)
	{
		debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Local: %s", (*i).localid.c_str());
		for (std::vector<std::string>::iterator j = (*i).remoteids.begin();
			j != (*i).remoteids.end(); j++)
		{
			debugprint(localdebugmode, PLUGIN_SHORT_NAME ": Remote: %s", (*j).c_str());
		}
	}
}

