/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2. */

#include "imspector.h"

#define PLUGIN_NAME "MSN IMSpector protocol plugin"
#define PROTOCOL_NAME "MSN"
#define PROTOCOL_PORT 1863

extern "C"
{
	bool initprotocolplugin(struct protocolplugininfo &pprotocolplugininfo,
		class Options &options, bool debugmode);
	void closeprotocolplugin(void);
	int processpacket(bool outgoing, class Socket &incomingsock, char *replybuffer, 
		int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress,
		std::vector<struct messageextent> &messageextents);
};

#pragma pack(2)

struct p2pheader
{
	uint32_t sessionid;
	uint32_t id;
	uint64_t offset;
	uint64_t datasize;
	uint32_t messagesize;
	uint32_t flags;
	uint32_t ackid;
	uint32_t ackuid;
	uint64_t acksize;
};

struct context
{
	uint32_t headerlength;
	uint32_t version;
	uint64_t filesize;
	uint32_t type;
	uint16_t filename[260];
};

#pragma pack()

void setremoteid(std::string id);
char *chopline(char* buffer, std::string &command, std::vector<std::string> &args, int &argc);
bool processmessage(bool outgoing, std::string id, int headerlength, char *msg, std::vector<struct imevent> &imevents,
	std::vector<struct messageextent> &messageextents);
char *getheadervalues(char *buffer, std::map<std::string, std::string> &headers);
char *getstring(char *buffer, std::string &str);

#define MSGTYPE_NULL 0
#define MSGTYPE_PLAIN 1
#define MSGTYPE_P2P 2
#define MSGTYPE_CONTROL 3

std::string localid = "Unknown";
std::string remoteid = "Unknown";
bool groupchat = false;
bool gotremoteid = false;
int packetcount = 0;
bool tracing = false;
std::map<uint32_t, std::string> filetransfers;
bool localdebugmode = false;

bool initprotocolplugin(struct protocolplugininfo &protocolplugininfo,
	class Options &options, bool debugmode)
{
	if (options["msn_protocol"] != "on") return false;

	localdebugmode = debugmode;
	
	protocolplugininfo.pluginname = PLUGIN_NAME;
	protocolplugininfo.protocolname = PROTOCOL_NAME;
	protocolplugininfo.port = htons(PROTOCOL_PORT);
	
	if (options["msn_trace"] == "on") tracing = true;
	
	return true;
}

void closeprotocolplugin(void)
{
	return;
}

/* The main plugin function. See protocolplugin.cpp. */
int processpacket(bool outgoing, class Socket &incomingsock, char *replybuffer, 
	int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress,
	std::vector<struct messageextent> &messageextents)
{
	char string[STRING_SIZE];
	memset(string, 0, STRING_SIZE);
	
	int headerlength;
	if ((headerlength = incomingsock.recvline(string, STRING_SIZE)) < 0) return 1;
	if (headerlength < 0) return 1;
	
	debugprint(localdebugmode, PROTOCOL_NAME ": Got %d bytes of header\n", headerlength);
	
	memcpy(replybuffer, string, headerlength);
	*replybufferlength = headerlength;
	
	std::string command;
	std::vector<std::string> args; int argc;

	char *s;
	s = chopline(string, command, args, argc);
	
	debugprint(localdebugmode, PROTOCOL_NAME ": Command: %s\n", command.c_str());
	
	/* These are all ways of getting the party (ID) information. */
	if (outgoing)
	{
		/* The local user is logging in. */
		if (command == "ANS" && argc > 1) 
			localid = args[1];
	}
	else
	{
		/* The local user completed the login. */
		if (command == "USR")
		{
			if (args[1] == "OK" && argc > 2)
				localid = args[2];
		}
		/* A remote user joined the chat. */
		if (command == "JOI" && argc > 0)
			setremoteid(args[0]);
		/* Could be multiple of these - group chats */
		if (command == "IRO" && argc > 3)
			setremoteid(args[3]);
	}
	
	if (command == "MSG" && argc > 2)
	{
		/* msgbuffer holds the actual message, ie. all data after the first line
		 * packet header. */
		char msgbuffer[BUFFER_SIZE];
		
		memset(msgbuffer, 0, BUFFER_SIZE);
		
		/* arg3 is the number of bytes following the end of the first line. */
		int lengthint = atol(args[2].c_str());

		if (!(incomingsock.recvalldata(msgbuffer, lengthint))) return 1;
		
		/* We get messages from "hotmail"  that we don't need to care about. */
		if (args[0] != "Hotmail")
			processmessage(outgoing, args[0], headerlength, msgbuffer, imevents, messageextents);
		
		/* Now we copy the msgbuffer back into the replybuffer, starting at the end
		 * of the first line. This gives us a chance to modify msgbuffer when we do
		 * some content filtering. */
		memcpy(replybuffer + headerlength, msgbuffer, lengthint);
		*replybufferlength += lengthint;
	}
	
	if ((command == "ADL" || command == "RML" || command == "UUN" ||
		command == "UBN" || command == "GCF" ||
		command == "UUX" || command == "UBX" ||
		command == "QRY" || command == "PAG" || command == "NOT")
		&& argc > 1)
	{
		int lengthint = 0;
		
		lengthint = atol(args[(argc - 1)].c_str());
			
		debugprint(localdebugmode, PROTOCOL_NAME ": %d bytes of %s data\n",
			lengthint, command.c_str());

		char databuffer[BUFFER_SIZE];
			
		memset(databuffer, 0, BUFFER_SIZE);
	
		if (!(incomingsock.recvalldata(databuffer, lengthint))) return 1;
					
		memcpy(replybuffer + headerlength, databuffer, lengthint);
		*replybufferlength += lengthint;
	}
	
	if (tracing)
	{
		char filename[STRING_SIZE];
		memset(filename, 0, STRING_SIZE);
		
		snprintf(filename, STRING_SIZE - 1, TRACE_DIR "/msn.%d.%d", getpid(), packetcount);

		int fd = -1;
		if ((fd = creat(filename, 0600)) > 0)
		{
			write(fd, replybuffer, *replybufferlength);
			close(fd);
		}
	}
	
	packetcount++;

	return 0;
}

/* Sets the remoteid, depending on wether or not this is the first remote id
 * spotted. */
void setremoteid(std::string id)
{
	time_t timestamp;

	timestamp = time(&timestamp);

	if (!gotremoteid)
	{
		remoteid = id;
		gotremoteid = true;
	}
	else if (!groupchat)
	{
		remoteid = "groupchat-" + stringprintf("%d", (int) timestamp);
		debugprint(localdebugmode, PROTOCOL_NAME ": Group chat, %s\n", remoteid.c_str());
		groupchat = true;
	}
}

/* Chops the heading first line into its command and arg list. Returns the
 * start of the following lines. */
char *chopline(char* buffer, std::string &command, std::vector<std::string> &args, int &argc)
{
	char *s = buffer;
	
	/* Copy the command, while the character looks valid. */
	for (; *s && *s != ' ' && *s != '\r' && *s != '\n'; s++)
		command.push_back(*s);
	s++;
	
	/* Same for the argument list. */
	argc = 0;
	while (*s && *s != '\r' && *s != '\n')
	{
		std::string arg;
		for (; *s && *s != ' ' && *s != '\r' && *s != '\n'; s++)
			arg.push_back(*s);
		s++;
		
		args.push_back(arg); argc++;
	}
	
	/* Need to advance s to the start of the next line, skipping over the eol chars. */
	for (; *s && (*s == '\r' || *s == '\n'); s++);
	
	return s;
}

bool processmessage(bool outgoing, std::string id, int headerlength, char *msg, std::vector<struct imevent> &imevents,
	std::vector<struct messageextent> &messageextents)
{
	std::map<std::string, std::string> headers;

	int msgtype = MSGTYPE_NULL;
		
	char *start = msg;
	start = getheadervalues(start, headers);
	
	const char *contenttype = headers["Content-Type"].c_str();
			
	if (strncmp(contenttype, "text/plain;", 11) == 0)
			msgtype = MSGTYPE_PLAIN;
	if (strcmp(contenttype, "application/x-msnmsgrp2p") == 0)
			msgtype = MSGTYPE_P2P;
	if (strcmp(contenttype, "text/x-msmsgscontrol") == 0)
			msgtype = MSGTYPE_CONTROL;
			
	if (msgtype != MSGTYPE_NULL)
	{
		struct imevent imevent;
		
		imevent.timestamp = time(NULL);
		imevent.protocolname = PROTOCOL_NAME;
		imevent.outgoing = outgoing;
		imevent.localid = localid;
		imevent.remoteid = remoteid;
		/* We are not yet commited to creating an event. */
		imevent.type = TYPE_NULL;
		imevent.filtered = false;
		
		if (msgtype == MSGTYPE_PLAIN)
		{
			imevent.type = TYPE_MSG;
			
			if (outgoing)
				imevent.eventdata = start;
			else if (!groupchat)
				imevent.eventdata = start;
			else
				imevent.eventdata = id + ": " + start;
			
			struct messageextent messageextent;
			
			messageextent.start = start - msg + headerlength;
			messageextent.length = -1; /* NULL terminated. */
			
			messageextents.push_back(messageextent);
		}
		if (msgtype == MSGTYPE_P2P)
		{
			debugprint(localdebugmode, PROTOCOL_NAME ": P2P");
			
			struct p2pheader p2pheader;
			
			memcpy(&p2pheader, start, sizeof(struct p2pheader));
		
			debugprint(localdebugmode, PROTOCOL_NAME ": sessionid: %u id: %u offset: %llu datasize: %llu messagesize: %u",
				p2pheader.sessionid, p2pheader.id, p2pheader.offset, p2pheader.datasize,
				p2pheader.messagesize);

			start += sizeof(struct p2pheader);
			
			/* if Session ID is 0 then it is a "start transfer" message. */
			if (!p2pheader.sessionid)
			{
				std::string invite;
				
				start = getstring(start, invite);
							
				if (strncmp(invite.c_str(), "INVITE ", 7) == 0)
				{
					debugprint(localdebugmode, PROTOCOL_NAME ": now onto header level two");
					
					std::map<std::string, std::string> headersleveltwo;
					start = getheadervalues(start, headersleveltwo);
					
					debugprint(localdebugmode, PROTOCOL_NAME ": now onto header level three");
					
					std::map<std::string, std::string> headerslevelthree;
					start = getheadervalues(start, headerslevelthree);
					
					/* AppID is 2 for normal file transfers. We will ignore buddy icons. */
					if (headerslevelthree["AppID"] == "2")
					{
						struct context context;
						
						memset(&context, 0, sizeof(struct context));
						
						decodebase64(headerslevelthree["Context"], (uint8_t *) &context, sizeof(struct context));
			
						debugprint(localdebugmode, PROTOCOL_NAME ": headerlength: %u version: %u filesize: %llu type: %u",
							context.headerlength, context.version, context.filesize, context.type);
						
						std::string filename;
						
						for (int c = 0; context.filename[c]; c++)
							filename += context.filename[c];					
						
						std::string sessionid = headerslevelthree["SessionID"];
						
						if (sessionid != "")
						{
							debugprint(localdebugmode, PROTOCOL_NAME ": FT sessionid: %s filename: %s",
								sessionid.c_str(), filename.c_str());
							filetransfers[atol(sessionid.c_str())] = filename;
						}
						
						char temp[STRING_SIZE];
						
						memset(temp, 0, STRING_SIZE);
						
						snprintf(temp, STRING_SIZE - 1, "%s %llu bytes", filename.c_str(), 
							(long long unsigned int)context.filesize);
						
						imevent.type = TYPE_FILE;		
						imevent.eventdata = temp;
					}
				}
			}
		}
		if (msgtype == MSGTYPE_CONTROL)
		{
			if (headers["TypingUser"] != "")
			{
				imevent.type = TYPE_TYPING;
				imevent.eventdata = "";
			}
		}
		
		if (imevent.type != TYPE_NULL)
		{
			std::transform(imevent.localid.begin(), imevent.localid.end(), imevent.localid.begin(), tolower);
			std::transform(imevent.remoteid.begin(), imevent.remoteid.end(), imevent.remoteid.begin(), tolower);
			imevents.push_back(imevent);
			
			return true;
		}
	}
	
	return false;
}

char *getheadervalues(char *buffer, std::map<std::string, std::string> &headers)
{
	char *s = buffer;
	
	while (*s && *s != '\r')
	{
		std::string header, value;
		
		while (*s && *s != ':')
		{
			header += *s;
			s++;
		}
		s++;
		while (*s && *s == ' ') s++;
		while (*s && *s != '\r')
		{
			value += *s;
			s++;
		}
		
		headers[header] = value;
		
		debugprint(localdebugmode, PROTOCOL_NAME ": header: %s value: %s", header.c_str(),
			value.c_str());
		
		if (!*s) break;
		
		s += 2;
		
		if (*s && *s == '\r') break;
	}

	s += 2;
	
	return s;
}

char *getstring(char *buffer, std::string &str)
{
	char *s = buffer;

	while (*s && *s != '\r')
	{
		str += *s;
		s++;
	}
	s += 2;
	
	return s;
}
